#ifndef __TYPE_H
#define __TYPE_H

#include <iostream>  //For the pretty-printer
#include <stdlib.h> //for exit()

#include <map>
#include <vector>

using namespace std; 

enum type_t
  { 
    INT,
    CHAR,
    POINTER,
    ARRAY,
    RECORD,

    ID, //Type identifier

    VOID //procedure return type
  };

class Type
{
 public:
  type_t type_id;

  Type* element_type; //array and pointer
  int size;           //array size
  map<string,Type*> record_field; //record

  string id; //type identifier

  /*
   * Constructors
   */

  Type(type_t base_type);
  //Base type
  Type(type_t base_type, Type* element_type); 
  //Pointer
  Type(type_t compound_type, Type* element_type, int size); 
  //Array
  Type(type_t compound_type, map<string,Type*> record_field);
  //Record

  Type(type_t base_type, char* identifier);
  //Identifier

  int is_base();
  //Base type? (INT, CHAR, POINTER)

  int is_well_formed();
  //Check for non-pointed recursion

  int operator==(Type& t);
  //Deep type equivalence

  Type* un_ref();                 //
  Type* un_index();               // used for type checking
  Type* un_field(char* field_id); //

  int get_field_rank(char* field_id);

  Type* clone();

  void print_dot(ostream& sout, string type_name);
  void show_dot(string file_name);

 private:
  int is_well_formed_(int pointer_reached, vector<Type*> path);
  void get_nodes(); //collect the nodes of 'this', used by print_dot
};

//pretty-printer
ostream& operator <<(ostream& sout, Type& e);

//private
string string_of_type_id(type_t type_id);

#endif
